use super::{
    panes::{Pane, PaneBehavior},
    shortcuts,
};

use egui::{Key, Modifiers};
use egui_tiles::{Behavior, Container, Linear, LinearDir, Tile, TileId, Tiles, Tree};

pub struct ComposableView {
    panes_tree: Tree<Pane>,
    behavior: ComposableBehavior,
}

// Implementing the default trait allows us to define a default configuration for our app
impl Default for ComposableView {
    fn default() -> Self {
        let mut tiles = Tiles::default();
        let root = tiles.insert_pane(Pane::default());
        let panes_tree = egui_tiles::Tree::new("my_tree", root, tiles);

        Self {
            panes_tree,
            behavior: Default::default(),
        }
    }
}

// An app must implement the `App` trait to define how the ui is built
impl eframe::App for ComposableView {
    // The update function is called each time the UI needs repainting!
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        // get the id of the hovered pane, in order to apply actions to it
        let hovered_pane = self
            .panes_tree
            .tiles
            .iter()
            .find(|(_, tile)| matches!(tile, Tile::Pane(pane) if pane.contains_pointer()))
            .map(|(id, _)| *id);

        // Capture any pane action generated by pane children
        let pane_action = self.behavior.action.take();
        let mut pane_action = pane_action.zip(hovered_pane);

        // Capture any pane action generated by keyboard shortcuts
        if let Some(hovered_pane) = hovered_pane {
            let key_action_pairs = [
                ((Modifiers::NONE, Key::V), PaneAction::SplitV),
                ((Modifiers::NONE, Key::H), PaneAction::SplitH),
                ((Modifiers::NONE, Key::C), PaneAction::Close),
            ];
            pane_action = pane_action.or(shortcuts::map_to_action(ctx, &key_action_pairs[..])
                .map(|action| (action, hovered_pane)));
        }

        // If an action was triggered, we consume it
        if let Some((action, hovered_tile)) = pane_action.take() {
            match action {
                PaneAction::SplitH => {
                    let hovered_tile_pane = self.panes_tree.tiles.remove(hovered_tile).unwrap();
                    let left_pane = self.panes_tree.tiles.insert_new(hovered_tile_pane);
                    let right_pane = self.panes_tree.tiles.insert_pane(Pane::default());
                    self.panes_tree.tiles.insert(
                        hovered_tile,
                        Tile::Container(Container::Linear(Linear::new_binary(
                            LinearDir::Horizontal,
                            [left_pane, right_pane],
                            0.5,
                        ))),
                    );
                }
                PaneAction::SplitV => {
                    let hovered_tile_pane = self.panes_tree.tiles.remove(hovered_tile).unwrap();
                    let replaced = self.panes_tree.tiles.insert_new(hovered_tile_pane);
                    let lower_pane = self.panes_tree.tiles.insert_pane(Pane::default());
                    self.panes_tree.tiles.insert(
                        hovered_tile,
                        Tile::Container(Container::Linear(Linear::new_binary(
                            LinearDir::Vertical,
                            [replaced, lower_pane],
                            0.5,
                        ))),
                    );
                }
                PaneAction::Close => {
                    // Ignore if the root pane is the only one
                    if self.panes_tree.tiles.len() != 1 {
                        self.panes_tree.remove_recursively(hovered_tile);
                    }
                }
                PaneAction::Replace(new_pane) => {
                    self.panes_tree
                        .tiles
                        .insert(hovered_tile, Tile::Pane(*new_pane));
                }
            }
        }

        // A central panel covers the remainder of the screen, i.e. whatever area is left after adding other panels.
        egui::CentralPanel::default().show(ctx, |ui| {
            self.panes_tree.ui(&mut self.behavior, ui);
        });

        // Show a panel at the bottom of the screen with few global controls
        egui::TopBottomPanel::bottom("bottom_control").show(ctx, |ui| {
            egui::global_theme_preference_switch(ui);
        });
    }
}

/// Behavior for the tree of panes in the composable view
#[derive(Default)]
pub struct ComposableBehavior {
    pub action: Option<PaneAction>,
}

impl Behavior<Pane> for ComposableBehavior {
    fn pane_ui(
        &mut self,
        ui: &mut egui::Ui,
        _tile_id: TileId,
        pane: &mut Pane,
    ) -> egui_tiles::UiResponse {
        let PaneResponse {
            action_called,
            drag_response,
        } = pane.ui(ui);
        // Capture the action and store it to be consumed in the update function
        if let Some(action_called) = action_called {
            self.action = Some(action_called);
        }
        drag_response
    }

    fn tab_title_for_pane(&mut self, _pane: &Pane) -> egui::WidgetText {
        "Tab".into()
    }
}

#[derive(Clone, Debug)]
pub struct PaneResponse {
    pub action_called: Option<PaneAction>,
    pub drag_response: egui_tiles::UiResponse,
}

impl PaneResponse {
    pub fn set_action(&mut self, action: PaneAction) {
        self.action_called = Some(action);
    }

    pub fn set_drag_started(&mut self) {
        self.drag_response = egui_tiles::UiResponse::DragStarted;
    }
}

impl Default for PaneResponse {
    fn default() -> Self {
        Self {
            action_called: None,
            drag_response: egui_tiles::UiResponse::None,
        }
    }
}

#[derive(Clone, Debug)]
pub enum PaneAction {
    SplitH,
    SplitV,
    Close,
    Replace(Box<Pane>),
}
